@@ -66,23 +66,25 @@ const { Pipe, constants: PipeConstants } = internalBinding('pipe_wrap');
 const { owner_symbol } = require('internal/async_hooks').symbols;
 const { isArrayBufferView } = require('internal/util/types');
 const { SecureContext: NativeSecureContext } = internalBinding('crypto');
-const { ConnResetException, codes } = require('internal/errors');
 const {
-  ERR_INVALID_ARG_TYPE,
-  ERR_INVALID_ARG_VALUE,
-  ERR_MULTIPLE_CALLBACK,
-  ERR_SOCKET_CLOSED,
-  ERR_TLS_ALPN_CALLBACK_INVALID_RESULT,
-  ERR_TLS_ALPN_CALLBACK_WITH_PROTOCOLS,
-  ERR_TLS_DH_PARAM_SIZE,
-  ERR_TLS_HANDSHAKE_TIMEOUT,
-  ERR_TLS_INVALID_CONTEXT,
-  ERR_TLS_RENEGOTIATION_DISABLED,
-  ERR_TLS_REQUIRED_SERVER_NAME,
-  ERR_TLS_SESSION_ATTACK,
-  ERR_TLS_SNI_FROM_SERVER,
-  ERR_TLS_INVALID_STATE,
-} = codes;
+  ConnResetException,
+  codes: {
+    ERR_INVALID_ARG_TYPE,
+    ERR_INVALID_ARG_VALUE,
+    ERR_MULTIPLE_CALLBACK,
+    ERR_SOCKET_CLOSED,
+    ERR_TLS_ALPN_CALLBACK_INVALID_RESULT,
+    ERR_TLS_ALPN_CALLBACK_WITH_PROTOCOLS,
+    ERR_TLS_DH_PARAM_SIZE,
+    ERR_TLS_HANDSHAKE_TIMEOUT,
+    ERR_TLS_INVALID_CONTEXT,
+    ERR_TLS_INVALID_STATE,
+    ERR_TLS_RENEGOTIATION_DISABLED,
+    ERR_TLS_REQUIRED_SERVER_NAME,
+    ERR_TLS_SESSION_ATTACK,
+    ERR_TLS_SNI_FROM_SERVER,
+  },
+} = require('internal/errors');
 const { onpskexchange: kOnPskExchange } = internalBinding('symbols');
 const {
   getOptionValue,
