@@ -1,22 +1,23 @@
 'use strict';
 
 const {
-  ArrayPrototypePush,
   ArrayPrototypePop,
+  ArrayPrototypePush,
   Error,
   ErrorCaptureStackTrace,
+  FunctionPrototypeBind,
   MathMax,
   MathMin,
   Promise,
   PromisePrototypeThen,
-  PromiseResolve,
   PromiseReject,
+  PromiseResolve,
   SafeArrayIterator,
   SafePromisePrototypeFinally,
   Symbol,
   SymbolAsyncDispose,
   Uint8Array,
-  FunctionPrototypeBind,
+  uncurryThis,
 } = primordials;
 
 const { fs: constants } = internalBinding('constants');
@@ -30,8 +31,12 @@ const {
 
 const binding = internalBinding('fs');
 const { Buffer } = require('buffer');
+const { isBuffer: BufferIsBuffer } = Buffer;
+const BufferToString = uncurryThis(Buffer.prototype.toString);
 
 const {
+  AbortError,
+  aggregateTwoErrors,
   codes: {
     ERR_ACCESS_DENIED,
     ERR_FS_FILE_TOO_LARGE,
@@ -39,8 +44,6 @@ const {
     ERR_INVALID_STATE,
     ERR_METHOD_NOT_IMPLEMENTED,
   },
-  AbortError,
-  aggregateTwoErrors,
 } = require('internal/errors');
 const { isArrayBufferView } = require('internal/util/types');
 
@@ -59,7 +62,6 @@ const {
   getStatFsFromBinding,
   getStatsFromBinding,
   getValidatedPath,
-  getValidMode,
   preprocessSymlinkDestination,
   stringToFlags,
   stringToSymlinkType,
@@ -90,10 +92,11 @@ const pathModule = require('path');
 const { isAbsolute } = pathModule;
 const { toPathIfFileURL } = require('internal/url');
 const {
+  emitExperimentalWarning,
+  getLazy,
   kEmptyObject,
   lazyDOMException,
   promisify,
-  getLazy,
 } = require('internal/util');
 const EventEmitter = require('events');
 const { StringDecoder } = require('string_decoder');
@@ -535,6 +538,7 @@ async function readFileHandle(filehandle, options) {
     throw new ERR_FS_FILE_TOO_LARGE(size);
 
   let totalRead = 0;
+  const noSize = size === 0;
   let buffer = Buffer.allocUnsafeSlow(length);
   let result = '';
   let offset = 0;
@@ -557,7 +561,7 @@ async function readFileHandle(filehandle, options) {
 
     if (bytesRead === 0 ||
         totalRead === size ||
-        (bytesRead !== buffer.length && !chunkedRead)) {
+        (bytesRead !== buffer.length && !chunkedRead && !noSize)) {
       const singleRead = bytesRead === totalRead;
 
       const bytesToCheck = chunkedRead ? totalRead : bytesRead;
@@ -567,7 +571,7 @@ async function readFileHandle(filehandle, options) {
       }
 
       if (!encoding) {
-        if (size === 0 && !singleRead) {
+        if (noSize && !singleRead) {
           ArrayPrototypePush(buffers, buffer);
           return Buffer.concat(buffers, totalRead);
         }
@@ -580,15 +584,17 @@ async function readFileHandle(filehandle, options) {
       result += decoder.end(buffer);
       return result;
     }
-
+    const readBuffer = bytesRead !== buffer.length ?
+      buffer.subarray(0, bytesRead) :
+      buffer;
     if (encoding) {
-      result += decoder.write(buffer);
+      result += decoder.write(readBuffer);
     } else if (size !== 0) {
       offset = totalRead;
     } else {
       buffers ??= [];
       // Unknown file size requires chunks.
-      ArrayPrototypePush(buffers, buffer);
+      ArrayPrototypePush(buffers, readBuffer);
       buffer = Buffer.allocUnsafeSlow(kReadFileUnknownBufferLength);
     }
   }
@@ -599,7 +605,6 @@ async function readFileHandle(filehandle, options) {
 async function access(path, mode = F_OK) {
   path = getValidatedPath(path);
 
-  mode = getValidMode(mode, 'access');
   return await PromisePrototypeThen(
     binding.access(pathModule.toNamespacedPath(path), mode, kUsePromises),
     undefined,
@@ -617,7 +622,6 @@ async function cp(src, dest, options) {
 async function copyFile(src, dest, mode) {
   src = getValidatedPath(src, 'src');
   dest = getValidatedPath(dest, 'dest');
-  mode = getValidMode(mode, 'copyFile');
   return await PromisePrototypeThen(
     binding.copyFile(pathModule.toNamespacedPath(src),
                      pathModule.toNamespacedPath(dest),
@@ -987,7 +991,11 @@ async function symlink(target, path, type_) {
   if (permission.isEnabled()) {
     // The permission model's security guarantees fall apart in the presence of
     // relative symbolic links. Thus, we have to prevent their creation.
-    if (typeof target !== 'string' || !isAbsolute(toPathIfFileURL(target))) {
+    if (BufferIsBuffer(target)) {
+      if (!isAbsolute(BufferToString(target))) {
+        throw new ERR_ACCESS_DENIED('relative symbolic link target');
+      }
+    } else if (typeof target !== 'string' || !isAbsolute(toPathIfFileURL(target))) {
       throw new ERR_ACCESS_DENIED('relative symbolic link target');
     }
   }
@@ -1267,11 +1275,19 @@ async function* _watch(filename, options = kEmptyObject) {
   yield* watch(filename, options);
 }
 
+const lazyGlob = getLazy(() => require('internal/fs/glob').Glob);
+async function* glob(pattern, options) {
+  emitExperimentalWarning('glob');
+  const Glob = lazyGlob();
+  yield* new Glob(pattern, options).glob();
+}
+
 module.exports = {
   exports: {
     access,
     copyFile,
     cp,
+    glob,
     open,
     opendir: promisify(opendir),
     rename,
