@@ -8,10 +8,12 @@ const {
   DatePrototypeGetTime,
   ErrorCaptureStackTrace,
   FunctionPrototypeCall,
-  Number,
-  NumberIsFinite,
   MathMin,
   MathRound,
+  Number,
+  NumberIsFinite,
+  ObjectDefineProperties,
+  ObjectDefineProperty,
   ObjectIs,
   ObjectSetPrototypeOf,
   ReflectApply,
@@ -22,13 +24,11 @@ const {
   Symbol,
   TypedArrayPrototypeAt,
   TypedArrayPrototypeIncludes,
-  uncurryThis,
 } = primordials;
 
-const permission = require('internal/process/permission');
-
 const { Buffer } = require('buffer');
 const {
+  UVException,
   codes: {
     ERR_FS_EISDIR,
     ERR_FS_INVALID_SYMLINK_TYPE,
@@ -38,7 +38,6 @@ const {
     ERR_OUT_OF_RANGE,
   },
   hideStackFrames,
-  UVException,
 } = require('internal/errors');
 const {
   isArrayBufferView,
@@ -49,6 +48,7 @@ const {
 const {
   kEmptyObject,
   once,
+  deprecate,
 } = require('internal/util');
 const { toPathIfFileURL } = require('internal/url');
 const {
@@ -65,8 +65,6 @@ const kType = Symbol('type');
 const kStats = Symbol('stats');
 const assert = require('internal/assert');
 
-const { encodeUtf8String } = internalBinding('encoding_binding');
-
 const {
   fs: {
     F_OK = 0,
@@ -451,6 +449,54 @@ function dateFromMs(ms) {
   return new Date(MathRound(Number(ms)));
 }
 
+const lazyDateFields = {
+  __proto__: null,
+  atime: {
+    __proto__: null,
+    enumerable: true,
+    configurable: true,
+    get() {
+      return this.atime = dateFromMs(this.atimeMs);
+    },
+    set(value) {
+      ObjectDefineProperty(this, 'atime', { __proto__: null, value, writable: true });
+    },
+  },
+  mtime: {
+    __proto__: null,
+    enumerable: true,
+    configurable: true,
+    get() {
+      return this.mtime = dateFromMs(this.mtimeMs);
+    },
+    set(value) {
+      ObjectDefineProperty(this, 'mtime', { __proto__: null, value, writable: true });
+    },
+  },
+  ctime: {
+    __proto__: null,
+    enumerable: true,
+    configurable: true,
+    get() {
+      return this.ctime = dateFromMs(this.ctimeMs);
+    },
+    set(value) {
+      ObjectDefineProperty(this, 'ctime', { __proto__: null, value, writable: true });
+    },
+  },
+  birthtime: {
+    __proto__: null,
+    enumerable: true,
+    configurable: true,
+    get() {
+      return this.birthtime = dateFromMs(this.birthtimeMs);
+    },
+    set(value) {
+      ObjectDefineProperty(this, 'birthtime', { __proto__: null, value, writable: true });
+    },
+  },
+};
+
 function BigIntStats(dev, mode, nlink, uid, gid, rdev, blksize,
                      ino, size, blocks,
                      atimeNs, mtimeNs, ctimeNs, birthtimeNs) {
@@ -465,14 +511,11 @@ function BigIntStats(dev, mode, nlink, uid, gid, rdev, blksize,
   this.mtimeNs = mtimeNs;
   this.ctimeNs = ctimeNs;
   this.birthtimeNs = birthtimeNs;
-  this.atime = dateFromMs(this.atimeMs);
-  this.mtime = dateFromMs(this.mtimeMs);
-  this.ctime = dateFromMs(this.ctimeMs);
-  this.birthtime = dateFromMs(this.birthtimeMs);
 }
 
 ObjectSetPrototypeOf(BigIntStats.prototype, StatsBase.prototype);
 ObjectSetPrototypeOf(BigIntStats, StatsBase);
+ObjectDefineProperties(BigIntStats.prototype, lazyDateFields);
 
 BigIntStats.prototype._checkModeProperty = function(property) {
   if (isWindows && (property === S_IFIFO || property === S_IFBLK ||
@@ -491,18 +534,11 @@ function Stats(dev, mode, nlink, uid, gid, rdev, blksize,
   this.mtimeMs = mtimeMs;
   this.ctimeMs = ctimeMs;
   this.birthtimeMs = birthtimeMs;
-  this.atime = dateFromMs(atimeMs);
-  this.mtime = dateFromMs(mtimeMs);
-  this.ctime = dateFromMs(ctimeMs);
-  this.birthtime = dateFromMs(birthtimeMs);
 }
 
 ObjectSetPrototypeOf(Stats.prototype, StatsBase.prototype);
 ObjectSetPrototypeOf(Stats, StatsBase);
-
-// HACK: Workaround for https://github.com/standard-things/esm/issues/821.
-// TODO(ronag): Remove this as soon as `esm` publishes a fixed version.
-Stats.prototype.isFile = StatsBase.prototype.isFile;
+ObjectDefineProperties(Stats.prototype, lazyDateFields);
 
 Stats.prototype._checkModeProperty = function(property) {
   if (isWindows && (property === S_IFIFO || property === S_IFBLK ||
@@ -695,31 +731,10 @@ const validatePath = hideStackFrames((path, propName = 'path') => {
   );
 });
 
-// TODO(rafaelgss): implement the path.resolve on C++ side
-// See: https://github.com/nodejs/node/pull/44004#discussion_r930958420
-// The permission model needs the absolute path for the fs_permission
-const resolvePath = pathModule.resolve;
-const { isBuffer: BufferIsBuffer, from: BufferFrom } = Buffer;
-const BufferToString = uncurryThis(Buffer.prototype.toString);
-function possiblyTransformPath(path) {
-  if (permission.isEnabled()) {
-    if (typeof path === 'string') {
-      return resolvePath(path);
-    }
-    assert(isUint8Array(path));
-    if (!BufferIsBuffer(path)) path = BufferFrom(path);
-    // Avoid Buffer.from() and use a C++ binding instead to encode the result
-    // of path.resolve() in order to prevent path traversal attacks that
-    // monkey-patch Buffer internals.
-    return encodeUtf8String(resolvePath(BufferToString(path)));
-  }
-  return path;
-}
-
 const getValidatedPath = hideStackFrames((fileURLOrPath, propName = 'path') => {
   const path = toPathIfFileURL(fileURLOrPath);
   validatePath(path, propName);
-  return possiblyTransformPath(path);
+  return path;
 });
 
 const getValidatedFd = hideStackFrames((fd, propName = 'fd') => {
@@ -946,22 +961,21 @@ module.exports = {
   BigIntStats,  // for testing
   copyObject,
   Dirent,
+  DirentFromStats,
   emitRecursiveRmdirWarning,
   getDirent,
   getDirents,
   getOptions,
   getValidatedFd,
   getValidatedPath,
-  getValidMode,
   handleErrorFromBinding,
-  possiblyTransformPath,
   preprocessSymlinkDestination,
   realpathCacheKey: Symbol('realpathCacheKey'),
   getStatFsFromBinding,
   getStatsFromBinding,
   stringToFlags,
   stringToSymlinkType,
-  Stats,
+  Stats: deprecate(Stats, 'fs.Stats constructor is deprecated.', 'DEP0180'),
   toUnixTimestamp,
   validateBufferArray,
   validateCpOptions,
