@@ -1,8 +1,5 @@
 'use strict';
 
-// This is needed to avoid cycles in esm/resolve <-> cjs/loader
-require('internal/modules/cjs/loader');
-
 const {
   ArrayPrototypeJoin,
   ArrayPrototypeMap,
@@ -15,21 +12,44 @@ const {
   hardenRegExp,
 } = primordials;
 
+
+// This is needed to avoid cycles in esm/resolve <-> cjs/loader
 const {
+  kIsExecuting,
+  kRequiredModuleSymbol,
+} = require('internal/modules/cjs/loader');
+
+const { imported_cjs_symbol } = internalBinding('symbols');
+
+const assert = require('internal/assert');
+const {
+  ERR_REQUIRE_CYCLE_MODULE,
   ERR_REQUIRE_ESM,
+  ERR_NETWORK_IMPORT_DISALLOWED,
   ERR_UNKNOWN_MODULE_FORMAT,
 } = require('internal/errors').codes;
 const { getOptionValue } = require('internal/options');
-const { pathToFileURL, isURL } = require('internal/url');
-const { emitExperimentalWarning } = require('internal/util');
+const { isURL, pathToFileURL, URL } = require('internal/url');
+const { emitExperimentalWarning, kEmptyObject } = require('internal/util');
 const {
+  compileSourceTextModule,
   getDefaultConditions,
 } = require('internal/modules/esm/utils');
+const { kImplicitTypeAttribute } = require('internal/modules/esm/assert');
+const { canParse } = internalBinding('url');
+const { ModuleWrap, kEvaluating, kEvaluated } = internalBinding('module_wrap');
+const {
+  urlToFilename,
+} = require('internal/modules/helpers');
 let defaultResolve, defaultLoad, defaultLoadSync, importMetaInitializer;
 
+/**
+ * @typedef {import('url').URL} URL
+ */
+
 /**
  * Lazy loads the module_map module and returns a new instance of ResolveCache.
- * @returns {import('./module_map.js').ResolveCache')}
+ * @returns {import('./module_map.js').ResolveCache}
  */
 function newResolveCache() {
   const { ResolveCache } = require('internal/modules/esm/module_map');
@@ -38,7 +58,7 @@ function newResolveCache() {
 
 /**
  * Generate a load cache (to store the final result of a load-chain for a particular module).
- * @returns {import('./module_map.js').LoadCache')}
+ * @returns {import('./module_map.js').LoadCache}
  */
 function newLoadCache() {
   const { LoadCache } = require('internal/modules/esm/module_map');
@@ -62,6 +82,10 @@ function getTranslators() {
  */
 let hooksProxy;
 
+/**
+ * @typedef {import('../cjs/loader.js').Module} CJSModule
+ */
+
 /**
  * @typedef {Record<string, any>} ModuleExports
  */
@@ -85,11 +109,6 @@ class ModuleLoader {
    */
   #defaultConditions = getDefaultConditions();
 
-  /**
-   * The index for assigning unique URLs to anonymous module evaluation
-   */
-  evalIndex = 0;
-
   /**
    * Registry of resolved specifiers
    */
@@ -187,33 +206,20 @@ class ModuleLoader {
     }
   }
 
-  async eval(
-    source,
-    url = pathToFileURL(`${process.cwd()}/[eval${++this.evalIndex}]`).href,
-  ) {
+  async eval(source, url, isEntryPoint = false) {
     const evalInstance = (url) => {
-      const { ModuleWrap } = internalBinding('module_wrap');
-      const { registerModule } = require('internal/modules/esm/utils');
-      const module = new ModuleWrap(url, undefined, source, 0, 0);
-      registerModule(module, {
-        __proto__: null,
-        initializeImportMeta: (meta, wrap) => this.importMetaInitialize(meta, { url }),
-        importModuleDynamically: (specifier, { url }, importAttributes) => {
-          return this.import(specifier, url, importAttributes);
-        },
-      });
-
-      return module;
+      return compileSourceTextModule(url, source, this);
     };
-    const ModuleJob = require('internal/modules/esm/module_job');
+    const { ModuleJob } = require('internal/modules/esm/module_job');
     const job = new ModuleJob(
       this, url, undefined, evalInstance, false, false);
     this.loadCache.set(url, undefined, job);
-    const { module } = await job.run();
+    const { module } = await job.run(isEntryPoint);
 
     return {
       __proto__: null,
       namespace: module.getNamespace(),
+      module,
     };
   }
 
@@ -257,6 +263,150 @@ class ModuleLoader {
     return job;
   }
 
+  /**
+   * This constructs (creates, instantiates and evaluates) a module graph that
+   * is require()'d.
+   * @param {CJSModule} mod CJS module wrapper of the ESM.
+   * @param {string} filename Resolved filename of the module being require()'d
+   * @param {string} source Source code. TODO(joyeecheung): pass the raw buffer.
+   * @param {string} isMain Whether this module is a main module.
+   * @param {CJSModule|undefined} parent Parent module, if any.
+   * @returns {{ModuleWrap}}
+   */
+  importSyncForRequire(mod, filename, source, isMain, parent) {
+    const url = pathToFileURL(filename).href;
+    let job = this.loadCache.get(url, kImplicitTypeAttribute);
+    // This module job is already created:
+    // 1. If it was loaded by `require()` before, at this point the instantiation
+    //    is already completed and we can check the whether it is in a cycle
+    //    (in that case the module status is kEvaluaing), and whether the
+    //    required graph is synchronous.
+    // 2. If it was loaded by `import` before, only allow it if it's already evaluated
+    //    to forbid cycles.
+    //    TODO(joyeecheung): ensure that imported synchronous graphs are evaluated
+    //    synchronously so that any previously imported synchronous graph is already
+    //    evaluated at this point.
+    if (job !== undefined) {
+      mod[kRequiredModuleSymbol] = job.module;
+      if (job.module.getStatus() !== kEvaluated) {
+        const parentFilename = urlToFilename(parent?.filename);
+        let message = `Cannot require() ES Module ${filename} in a cycle.`;
+        if (parentFilename) {
+          message += ` (from ${parentFilename})`;
+        }
+        throw new ERR_REQUIRE_CYCLE_MODULE(message);
+      }
+      return job.module.getNamespaceSync();
+    }
+    // TODO(joyeecheung): refactor this so that we pre-parse in C++ and hit the
+    // cache here, or use a carrier object to carry the compiled module script
+    // into the constructor to ensure cache hit.
+    const wrap = compileSourceTextModule(url, source, this);
+    const inspectBrk = (isMain && getOptionValue('--inspect-brk'));
+
+    const { ModuleJobSync } = require('internal/modules/esm/module_job');
+    job = new ModuleJobSync(this, url, kEmptyObject, wrap, isMain, inspectBrk);
+    this.loadCache.set(url, kImplicitTypeAttribute, job);
+    mod[kRequiredModuleSymbol] = job.module;
+    return job.runSync().namespace;
+  }
+
+  /**
+   * Resolve individual module requests and create or get the cached ModuleWraps for
+   * each of them. This is only used to create a module graph being require()'d.
+   * @param {string} specifier Specifier of the the imported module.
+   * @param {string} parentURL Where the import comes from.
+   * @param {object} importAttributes import attributes from the import statement.
+   * @returns {ModuleJobBase}
+   */
+  getModuleJobForRequire(specifier, parentURL, importAttributes) {
+    assert(getOptionValue('--experimental-require-module'));
+
+    if (canParse(specifier)) {
+      const protocol = new URL(specifier).protocol;
+      if (protocol === 'https:' || protocol === 'http:') {
+        throw new ERR_NETWORK_IMPORT_DISALLOWED(specifier, parentURL,
+                                                'ES modules cannot be loaded by require() from the network');
+      }
+      assert(protocol === 'file:' || protocol === 'node:' || protocol === 'data:');
+    }
+
+    const requestKey = this.#resolveCache.serializeKey(specifier, importAttributes);
+    let resolveResult = this.#resolveCache.get(requestKey, parentURL);
+    if (resolveResult == null) {
+      resolveResult = this.defaultResolve(specifier, parentURL, importAttributes);
+      this.#resolveCache.set(requestKey, parentURL, resolveResult);
+    }
+
+    const { url, format } = resolveResult;
+    const resolvedImportAttributes = resolveResult.importAttributes ?? importAttributes;
+    let job = this.loadCache.get(url, resolvedImportAttributes.type);
+    if (job !== undefined) {
+      // This module is being evaluated, which means it's imported in a previous link
+      // in a cycle.
+      if (job.module.getStatus() === kEvaluating) {
+        const parentFilename = urlToFilename(parentURL);
+        let message = `Cannot import Module ${specifier} in a cycle.`;
+        if (parentFilename) {
+          message += ` (from ${parentFilename})`;
+        }
+        throw new ERR_REQUIRE_CYCLE_MODULE(message);
+      }
+      // Otherwise the module could be imported before but the evaluation may be already
+      // completed (e.g. the require call is lazy) so it's okay. We will return the
+      // module now and check asynchronicity of the entire graph later, after the
+      // graph is instantiated.
+      return job;
+    }
+
+    defaultLoadSync ??= require('internal/modules/esm/load').defaultLoadSync;
+    const loadResult = defaultLoadSync(url, { format, importAttributes });
+    const {
+      format: finalFormat,
+      responseURL,
+      source,
+    } = loadResult;
+
+    this.validateLoadResult(url, finalFormat);
+    if (finalFormat === 'wasm') {
+      assert.fail('WASM is currently unsupported by require(esm)');
+    }
+
+    const translator = getTranslators().get(finalFormat);
+    if (!translator) {
+      throw new ERR_UNKNOWN_MODULE_FORMAT(finalFormat, responseURL);
+    }
+
+    const isMain = (parentURL === undefined);
+    const wrap = FunctionPrototypeCall(translator, this, responseURL, source, isMain);
+    assert(wrap instanceof ModuleWrap);  // No asynchronous translators should be called.
+
+    if (process.env.WATCH_REPORT_DEPENDENCIES && process.send) {
+      process.send({ 'watch:import': [url] });
+    }
+
+    const cjsModule = wrap[imported_cjs_symbol];
+    if (cjsModule) {
+      assert(finalFormat === 'commonjs-sync');
+      // Check if the ESM initiating import CJS is being required by the same CJS module.
+      if (cjsModule && cjsModule[kIsExecuting]) {
+        const parentFilename = urlToFilename(parentURL);
+        let message = `Cannot import CommonJS Module ${specifier} in a cycle.`;
+        if (parentFilename) {
+          message += ` (from ${parentFilename})`;
+        }
+        throw new ERR_REQUIRE_CYCLE_MODULE(message);
+      }
+    }
+
+    const inspectBrk = (isMain && getOptionValue('--inspect-brk'));
+    const { ModuleJobSync } = require('internal/modules/esm/module_job');
+    job = new ModuleJobSync(this, url, importAttributes, wrap, isMain, inspectBrk);
+
+    this.loadCache.set(url, importAttributes.type, job);
+    return job;
+  }
+
   /**
    * Create and cache an object representing a loaded module.
    * @param {string} url The absolute URL that was resolved for this module
@@ -284,8 +434,9 @@ class ModuleLoader {
       (url, isMain) => callTranslator(this.loadSync(url, context), isMain) :
       async (url, isMain) => callTranslator(await this.load(url, context), isMain);
 
+    const isMain = parentURL === undefined;
     const inspectBrk = (
-      parentURL === undefined &&
+      isMain &&
       getOptionValue('--inspect-brk')
     );
 
@@ -293,13 +444,13 @@ class ModuleLoader {
       process.send({ 'watch:import': [url] });
     }
 
-    const ModuleJob = require('internal/modules/esm/module_job');
+    const { ModuleJob } = require('internal/modules/esm/module_job');
     const job = new ModuleJob(
       this,
       url,
       importAttributes,
       moduleProvider,
-      parentURL === undefined,
+      isMain,
       inspectBrk,
       sync,
     );
@@ -318,9 +469,9 @@ class ModuleLoader {
    *                                                  module import.
    * @returns {Promise<ModuleExports>}
    */
-  async import(specifier, parentURL, importAttributes) {
+  async import(specifier, parentURL, importAttributes, isEntryPoint = false) {
     const moduleJob = await this.getModuleJob(specifier, parentURL, importAttributes);
-    const { module } = await moduleJob.run();
+    const { module } = await moduleJob.run(isEntryPoint);
     return module.getNamespace();
   }
 
@@ -568,13 +719,30 @@ function getHooksProxy() {
   return hooksProxy;
 }
 
+let cascadedLoader;
+
+/**
+ * This is a singleton ESM loader that integrates the loader hooks, if any.
+ * It it used by other internal built-ins when they need to load ESM code
+ * while also respecting hooks.
+ * When built-ins need access to this loader, they should do
+ * require('internal/module/esm/loader').getOrInitializeCascadedLoader()
+ * lazily only right before the loader is actually needed, and don't do it
+ * in the top-level, to avoid circular dependencies.
+ * @returns {ModuleLoader}
+ */
+function getOrInitializeCascadedLoader() {
+  cascadedLoader ??= createModuleLoader();
+  return cascadedLoader;
+}
+
 /**
  * Register a single loader programmatically.
- * @param {string|import('url').URL} specifier
- * @param {string|import('url').URL} [parentURL] Base to use when resolving `specifier`; optional if
+ * @param {string|URL} specifier
+ * @param {string|URL} [parentURL] Base to use when resolving `specifier`; optional if
  * `specifier` is absolute. Same as `options.parentUrl`, just inline
  * @param {object} [options] Additional options to apply, described below.
- * @param {string|import('url').URL} [options.parentURL] Base to use when resolving `specifier`
+ * @param {string|URL} [options.parentURL] Base to use when resolving `specifier`
  * @param {any} [options.data] Arbitrary data passed to the loader's `initialize` hook
  * @param {any[]} [options.transferList] Objects in `data` that are changing ownership
  * @returns {void} We want to reserve the return value for potential future extension of the API.
@@ -598,12 +766,11 @@ function getHooksProxy() {
  * ```
  */
 function register(specifier, parentURL = undefined, options) {
-  const moduleLoader = require('internal/process/esm_loader').esmLoader;
   if (parentURL != null && typeof parentURL === 'object' && !isURL(parentURL)) {
     options = parentURL;
     parentURL = options.parentURL;
   }
-  moduleLoader.register(
+  getOrInitializeCascadedLoader().register(
     specifier,
     parentURL ?? 'data:',
     options?.data,
@@ -614,5 +781,6 @@ function register(specifier, parentURL = undefined, options) {
 module.exports = {
   createModuleLoader,
   getHooksProxy,
+  getOrInitializeCascadedLoader,
   register,
 };
