@@ -4,9 +4,8 @@ const {
   ArrayPrototypeMap,
   Boolean,
   JSONParse,
-  ObjectGetPrototypeOf,
-  ObjectPrototypeHasOwnProperty,
   ObjectKeys,
+  ObjectPrototypeHasOwnProperty,
   ReflectApply,
   SafeArrayIterator,
   SafeMap,
@@ -15,8 +14,6 @@ const {
   StringPrototypeReplaceAll,
   StringPrototypeSlice,
   StringPrototypeStartsWith,
-  Symbol,
-  SyntaxErrorPrototype,
   globalThis: { WebAssembly },
 } = primordials;
 
@@ -30,7 +27,10 @@ function lazyTypes() {
   return _TYPES = require('internal/util/types');
 }
 
-const { containsModuleSyntax } = internalBinding('contextify');
+const {
+  compileFunctionForCJSLoader,
+} = internalBinding('contextify');
+
 const { BuiltinModule } = require('internal/bootstrap/realm');
 const assert = require('internal/assert');
 const { readFileSync } = require('fs');
@@ -38,11 +38,15 @@ const { dirname, extname, isAbsolute } = require('path');
 const {
   loadBuiltinModule,
   stripBOM,
+  urlToFilename,
 } = require('internal/modules/helpers');
 const {
+  kIsCachedByESMLoader,
   Module: CJSModule,
-  cjsParseCache,
-  cjsExportsCache,
+  wrapModuleLoad,
+  kModuleSource,
+  kModuleExport,
+  kModuleExportNames,
 } = require('internal/modules/cjs/loader');
 const { fileURLToPath, pathToFileURL, URL } = require('internal/url');
 let debug = require('internal/util/debuglog').debuglog('esm', (fn) => {
@@ -56,9 +60,6 @@ const {
 const { maybeCacheSourceMap } = require('internal/source_map/source_map_cache');
 const moduleWrap = internalBinding('module_wrap');
 const { ModuleWrap } = moduleWrap;
-const asyncESM = require('internal/process/esm_loader');
-const { emitWarningSync } = require('internal/process/warning');
-const { internalCompileFunction } = require('internal/vm');
 
 // Lazy-loading to avoid circular dependencies.
 let getSourceSync;
@@ -80,7 +81,7 @@ let cjsParse;
  */
 async function initCJSParse() {
   if (typeof WebAssembly === 'undefined') {
-    cjsParse = require('internal/deps/cjs-module-lexer/lexer').parse;
+    initCJSParseSync();
   } else {
     const { parse, init } =
         require('internal/deps/cjs-module-lexer/dist/lexer');
@@ -88,14 +89,21 @@ async function initCJSParse() {
       await init();
       cjsParse = parse;
     } catch {
-      cjsParse = require('internal/deps/cjs-module-lexer/lexer').parse;
+      initCJSParseSync();
     }
   }
 }
 
+function initCJSParseSync() {
+  // TODO(joyeecheung): implement a binding that directly compiles using
+  // v8::WasmModuleObject::Compile() synchronously.
+  if (cjsParse === undefined) {
+    cjsParse = require('internal/deps/cjs-module-lexer/lexer').parse;
+  }
+}
+
 const translators = new SafeMap();
 exports.translators = translators;
-exports.enrichCJSError = enrichCJSError;
 
 let DECODER = null;
 /**
@@ -147,53 +155,16 @@ function errPath(url) {
   return url;
 }
 
-/**
- * Dynamically imports a module using the ESM loader.
- * @param {string} specifier - The module specifier to import.
- * @param {object} options - An object containing options for the import.
- * @param {string} options.url - The URL of the module requesting the import.
- * @param {Record<string, string>} [attributes] - An object containing attributes for the import.
- * @returns {Promise<import('internal/modules/esm/loader.js').ModuleExports>} The imported module.
- */
-async function importModuleDynamically(specifier, { url }, attributes) {
-  return asyncESM.esmLoader.import(specifier, url, attributes);
-}
-
 // Strategy for loading a standard JavaScript module.
-translators.set('module', async function moduleStrategy(url, source, isMain) {
+translators.set('module', function moduleStrategy(url, source, isMain) {
   assertBufferSource(source, true, 'load');
   source = stringify(source);
-  maybeCacheSourceMap(url, source);
   debug(`Translating StandardModule ${url}`);
-  const module = new ModuleWrap(url, undefined, source, 0, 0);
-  const { registerModule } = require('internal/modules/esm/utils');
-  registerModule(module, {
-    __proto__: null,
-    initializeImportMeta: (meta, wrap) => this.importMetaInitialize(meta, { url }),
-    importModuleDynamically,
-  });
+  const { compileSourceTextModule } = require('internal/modules/esm/utils');
+  const module = compileSourceTextModule(url, source, this);
   return module;
 });
 
-/**
- * Provide a more informative error for CommonJS imports.
- * @param {Error | any} err
- * @param {string} [content] Content of the file, if known.
- * @param {string} [filename] The filename of the erroring module.
- */
-function enrichCJSError(err, content, filename) {
-  if (err != null && ObjectGetPrototypeOf(err) === SyntaxErrorPrototype &&
-      containsModuleSyntax(content, filename)) {
-    // Emit the warning synchronously because we are in the middle of handling
-    // a SyntaxError that will throw and likely terminate the process before an
-    // asynchronous warning would be emitted.
-    emitWarningSync(
-      'To load an ES module, set "type": "module" in the package.json or use ' +
-      'the .mjs extension.',
-    );
-  }
-}
-
 /**
  * Loads a CommonJS module via the ESM Loader sync CommonJS translator.
  * This translator creates its own version of the `require` function passed into CommonJS modules.
@@ -203,37 +174,18 @@ function enrichCJSError(err, content, filename) {
  * @param {string} source - The source code of the module.
  * @param {string} url - The URL of the module.
  * @param {string} filename - The filename of the module.
+ * @param {boolean} isMain - Whether the module is the entrypoint
  */
-function loadCJSModule(module, source, url, filename) {
-  let compiledWrapper;
-  async function importModuleDynamically(specifier, _, importAttributes) {
-    return asyncESM.esmLoader.import(specifier, url, importAttributes);
-  }
-  try {
-    compiledWrapper = internalCompileFunction(
-      source,                         // code,
-      filename,                       // filename
-      0,                              // lineOffset
-      0,                              // columnOffset,
-      undefined,                      // cachedData
-      false,                          // produceCachedData
-      undefined,                      // parsingContext
-      undefined,                      // contextExtensions
-      [                               // params
-        'exports',
-        'require',
-        'module',
-        '__filename',
-        '__dirname',
-      ],
-      Symbol(`cjs:${filename}`),      // hostDefinedOptionsId
-      importModuleDynamically,        // importModuleDynamically
-    ).function;
-  } catch (err) {
-    enrichCJSError(err, source, filename);
-    throw err;
+function loadCJSModule(module, source, url, filename, isMain) {
+  const compileResult = compileFunctionForCJSLoader(source, filename, isMain, false);
+
+  const { function: compiledWrapper, sourceMapURL } = compileResult;
+  // Cache the source map for the cjs module if present.
+  if (sourceMapURL) {
+    maybeCacheSourceMap(url, source, module, false, undefined, sourceMapURL);
   }
 
+  const cascadedLoader = require('internal/modules/esm/loader').getOrInitializeCascadedLoader();
   const __dirname = dirname(filename);
   // eslint-disable-next-line func-name-matching,func-style
   const requireFn = function require(specifier) {
@@ -246,13 +198,13 @@ function loadCJSModule(module, source, url, filename) {
           importAttributes = { __proto__: null, type: 'json' };
           break;
         case '.node':
-          return CJSModule._load(specifier, module);
+          return wrapModuleLoad(specifier, module);
         default:
             // fall through
       }
       specifier = `${pathToFileURL(path)}`;
     }
-    const job = asyncESM.esmLoader.getModuleJobSync(specifier, url, importAttributes);
+    const job = cascadedLoader.getModuleJobSync(specifier, url, importAttributes);
     job.runSync();
     return cjsCache.get(job.url).exports;
   };
@@ -263,8 +215,8 @@ function loadCJSModule(module, source, url, filename) {
         specifier = `${pathToFileURL(path)}`;
       }
     }
-    const { url: resolvedURL } = asyncESM.esmLoader.resolveSync(specifier, url, kEmptyObject);
-    return StringPrototypeStartsWith(resolvedURL, 'file://') ? fileURLToPath(resolvedURL) : resolvedURL;
+    const { url: resolvedURL } = cascadedLoader.resolveSync(specifier, url, kEmptyObject);
+    return urlToFilename(resolvedURL);
   });
   setOwnProperty(requireFn, 'main', process.mainModule);
 
@@ -286,12 +238,10 @@ const cjsCache = new SafeMap();
 function createCJSModuleWrap(url, source, isMain, loadCJS = loadCJSModule) {
   debug(`Translating CJSModule ${url}`);
 
-  const filename = StringPrototypeStartsWith(url, 'file://') ? fileURLToPath(url) : url;
+  const filename = urlToFilename(url);
   // In case the source was not provided by the `load` step, we need fetch it now.
   source = stringify(source ?? getSource(new URL(url)).source);
 
-  maybeCacheSourceMap(url, source);
-
   const { exportNames, module } = cjsPreparseModuleExports(filename, source);
   cjsCache.set(url, module);
   const namesWithDefault = exportNames.has('default') ?
@@ -305,13 +255,13 @@ function createCJSModuleWrap(url, source, isMain, loadCJS = loadCJSModule) {
     debug(`Loading CJSModule ${url}`);
 
     if (!module.loaded) {
-      loadCJS(module, source, url, filename);
+      loadCJS(module, source, url, filename, !!isMain);
     }
 
     let exports;
-    if (cjsExportsCache.has(module)) {
-      exports = cjsExportsCache.get(module);
-      cjsExportsCache.delete(module);
+    if (module[kModuleExport] !== undefined) {
+      exports = module[kModuleExport];
+      module[kModuleExport] = undefined;
     } else {
       ({ exports } = module);
     }
@@ -330,10 +280,20 @@ function createCJSModuleWrap(url, source, isMain, loadCJS = loadCJSModule) {
       this.setExport(exportName, value);
     }
     this.setExport('default', exports);
-  });
-
+  }, module);
 }
 
+translators.set('commonjs-sync', function requireCommonJS(url, source, isMain) {
+  initCJSParseSync();
+  assert(!isMain);  // This is only used by imported CJS modules.
+
+  return createCJSModuleWrap(url, source, isMain, (module, source, url, filename, isMain) => {
+    assert(module === CJSModule._cache[filename]);
+    assert(!isMain);
+    wrapModuleLoad(filename, null, isMain);
+  });
+});
+
 // Handle CommonJS modules referenced by `require` calls.
 // This translator function must be sync, as `require` is sync.
 translators.set('require-commonjs', (url, source, isMain) => {
@@ -353,14 +313,9 @@ translators.set('commonjs', async function commonjsStrategy(url, source,
   // For backward-compatibility, it's possible to return a nullish value for
   // CJS source associated with a file: URL. In this case, the source is
   // obtained by calling the monkey-patchable CJS loader.
-  const cjsLoader = source == null ? (module, source, url, filename) => {
-    try {
-      assert(module === CJSModule._cache[filename]);
-      CJSModule._load(filename);
-    } catch (err) {
-      enrichCJSError(err, source, filename);
-      throw err;
-    }
+  const cjsLoader = source == null ? (module, source, url, filename, isMain) => {
+    assert(module === CJSModule._cache[filename]);
+    wrapModuleLoad(filename, undefined, isMain);
   } : loadCJSModule;
 
   try {
@@ -381,17 +336,16 @@ translators.set('commonjs', async function commonjsStrategy(url, source,
 function cjsPreparseModuleExports(filename, source) {
   // TODO: Do we want to keep hitting the user mutable CJS loader here?
   let module = CJSModule._cache[filename];
-  if (module) {
-    const cached = cjsParseCache.get(module);
-    if (cached) {
-      return { module, exportNames: cached.exportNames };
-    }
+  if (module && module[kModuleExportNames] !== undefined) {
+    return { module, exportNames: module[kModuleExportNames] };
   }
   const loaded = Boolean(module);
   if (!loaded) {
     module = new CJSModule(filename);
     module.filename = filename;
     module.paths = CJSModule._nodeModulePaths(module.path);
+    module[kIsCachedByESMLoader] = true;
+    module[kModuleSource] = source;
     CJSModule._cache[filename] = module;
   }
 
@@ -406,7 +360,7 @@ function cjsPreparseModuleExports(filename, source) {
   const exportNames = new SafeSet(new SafeArrayIterator(exports));
 
   // Set first for cycles.
-  cjsParseCache.set(module, { source, exportNames });
+  module[kModuleExportNames] = exportNames;
 
   if (reexports.length) {
     module.filename = filename;
@@ -533,6 +487,8 @@ translators.set('wasm', async function(url, source) {
 
   let compiled;
   try {
+    // TODO(joyeecheung): implement a binding that directly compiles using
+    // v8::WasmModuleObject::Compile() synchronously.
     compiled = await WebAssembly.compile(source);
   } catch (err) {
     err.message = errPath(url) + ': ' + err.message;
