@@ -29,8 +29,8 @@ const {
   ArrayPrototypeSlice,
   Error,
   FunctionPrototypeSymbolHasInstance,
-  ObjectDefineProperty,
   ObjectDefineProperties,
+  ObjectDefineProperty,
   ObjectSetPrototypeOf,
   StringPrototypeToLowerCase,
   Symbol,
@@ -252,8 +252,8 @@ ObjectDefineProperties(WritableState.prototype, {
     enumerable: false,
     get() { return (this[kState] & kWriteCb) !== 0 ? this[kWriteCbValue] : nop; },
     set(value) {
+      this[kWriteCbValue] = value;
       if (value) {
-        this[kWriteCbValue] = value;
         this[kState] |= kWriteCb;
       } else {
         this[kState] &= ~kWriteCb;
@@ -268,8 +268,8 @@ ObjectDefineProperties(WritableState.prototype, {
     enumerable: false,
     get() { return (this[kState] & kAfterWriteTickInfo) !== 0 ? this[kAfterWriteTickInfoValue] : null; },
     set(value) {
+      this[kAfterWriteTickInfoValue] = value;
       if (value) {
-        this[kAfterWriteTickInfoValue] = value;
         this[kState] |= kAfterWriteTickInfo;
       } else {
         this[kState] &= ~kAfterWriteTickInfo;
@@ -467,12 +467,12 @@ function _write(stream, chunk, encoding, cb) {
       }
     } else if (chunk instanceof Buffer) {
       encoding = 'buffer';
-    } else if (Stream._isUint8Array(chunk)) {
+    } else if (Stream._isArrayBufferView(chunk)) {
       chunk = Stream._uint8ArrayToBuffer(chunk);
       encoding = 'buffer';
     } else {
       throw new ERR_INVALID_ARG_TYPE(
-        'chunk', ['string', 'Buffer', 'Uint8Array'], chunk);
+        'chunk', ['string', 'Buffer', 'TypedArray', 'DataView'], chunk);
     }
   }
 
@@ -615,7 +615,8 @@ function onwrite(stream, er) {
   const sync = (state[kState] & kSync) !== 0;
   const cb = (state[kState] & kWriteCb) !== 0 ? state[kWriteCbValue] : nop;
 
-  state[kState] &= ~(kWriting | kExpectWriteCb | kWriteCb);
+  state.writecb = null;
+  state[kState] &= ~(kWriting | kExpectWriteCb);
   state.length -= state.writelen;
   state.writelen = 0;
 
@@ -876,7 +877,7 @@ function needFinish(state) {
 
 function onFinish(stream, state, err) {
   if ((state[kState] & kPrefinished) !== 0) {
-    errorOrDestroy(stream, err ?? ERR_MULTIPLE_CALLBACK());
+    errorOrDestroy(stream, err ?? new ERR_MULTIPLE_CALLBACK());
     return;
   }
   state.pendingcb--;
@@ -1004,8 +1005,8 @@ ObjectDefineProperties(Writable.prototype, {
       // where the writable side was disabled upon construction.
       // Compat. The user might manually disable writable side through
       // deprecated setter.
-      return !!w && w.writable !== false && !w.errored &&
-        (w[kState] & (kEnding | kEnded | kDestroyed)) === 0;
+      return !!w && w.writable !== false &&
+        (w[kState] & (kEnding | kEnded | kDestroyed | kErrored)) === 0;
     },
     set(val) {
       // Backwards compatible.
