@@ -1,9 +1,5 @@
 'use strict';
 
-const {
-  emitExperimentalWarning,
-} = require('internal/util');
-
 const {
   ArrayPrototypeAt,
   ArrayPrototypeForEach,
@@ -13,26 +9,33 @@ const {
   FunctionPrototypeApply,
   FunctionPrototypeBind,
   FunctionPrototypeToString,
-  globalThis,
   NumberIsNaN,
-  ObjectDefineProperty,
   ObjectDefineProperties,
+  ObjectDefineProperty,
   ObjectGetOwnPropertyDescriptor,
   ObjectGetOwnPropertyDescriptors,
   Promise,
   Symbol,
   SymbolAsyncIterator,
   SymbolDispose,
+  globalThis,
 } = primordials;
+
 const {
   validateAbortSignal,
   validateArray,
   validateNumber,
 } = require('internal/validators');
 
+const {
+  emitExperimentalWarning,
+} = require('internal/util');
 const {
   AbortError,
-  codes: { ERR_INVALID_STATE, ERR_INVALID_ARG_VALUE },
+  codes: {
+    ERR_INVALID_ARG_VALUE,
+    ERR_INVALID_STATE,
+  },
 } = require('internal/errors');
 
 const PriorityQueue = require('internal/priority_queue');
@@ -67,6 +70,32 @@ const TIMERS_DEFAULT_INTERVAL = {
   setImmediate: -1,
 };
 
+class Timeout {
+  constructor(opts) {
+    this.id = opts.id;
+    this.callback = opts.callback;
+    this.runAt = opts.runAt;
+    this.interval = opts.interval;
+    this.args = opts.args;
+  }
+
+  hasRef() {
+    return true;
+  }
+
+  ref() {
+    return this;
+  }
+
+  unref() {
+    return this;
+  }
+
+  refresh() {
+    return this;
+  }
+}
+
 class MockTimers {
   #realSetTimeout;
   #realClearTimeout;
@@ -260,7 +289,7 @@ class MockTimers {
 
   #createTimer(isInterval, callback, delay, ...args) {
     const timerId = this.#currentTimer++;
-    const timer = {
+    const opts = {
       __proto__: null,
       id: timerId,
       callback,
@@ -268,6 +297,8 @@ class MockTimers {
       interval: isInterval ? delay : undefined,
       args,
     };
+
+    const timer = new Timeout(opts);
     this.#executionQueue.insert(timer);
     return timer;
   }
@@ -622,8 +653,12 @@ class MockTimers {
       if (timer.runAt > this.#now) break;
       FunctionPrototypeApply(timer.callback, undefined, timer.args);
 
-      this.#executionQueue.shift();
-      timer.priorityQueuePosition = undefined;
+      // Check if the timeout was cleared by calling clearTimeout inside its own callback
+      const afterCallback = this.#executionQueue.peek();
+      if (afterCallback?.id === timer.id) {
+        this.#executionQueue.shift();
+        timer.priorityQueuePosition = undefined;
+      }
 
       if (timer.interval !== undefined) {
         timer.runAt += timer.interval;
