@@ -12,10 +12,10 @@ const {
   FunctionPrototypeCall,
   ObjectDefineProperties,
   ObjectDefineProperty,
+  ObjectFreeze,
   ObjectGetOwnPropertyDescriptor,
   ObjectGetOwnPropertyDescriptors,
   ObjectGetPrototypeOf,
-  ObjectFreeze,
   ObjectPrototypeHasOwnProperty,
   ObjectSetPrototypeOf,
   ObjectValues,
@@ -28,13 +28,14 @@ const {
   RegExpPrototypeGetHasIndices,
   RegExpPrototypeGetIgnoreCase,
   RegExpPrototypeGetMultiline,
+  RegExpPrototypeGetSource,
   RegExpPrototypeGetSticky,
   RegExpPrototypeGetUnicode,
-  RegExpPrototypeGetSource,
   SafeMap,
   SafeSet,
   SafeWeakMap,
   SafeWeakRef,
+  StringPrototypeIncludes,
   StringPrototypeReplace,
   StringPrototypeToLowerCase,
   StringPrototypeToUpperCase,
@@ -50,8 +51,8 @@ const {
     ERR_UNKNOWN_SIGNAL,
   },
   isErrorStackTraceLimitWritable,
-  uvErrmapGet,
   overrideStackTrace,
+  uvErrmapGet,
 } = require('internal/errors');
 const { signals } = internalBinding('constants').os;
 const {
@@ -476,18 +477,18 @@ function spliceOne(list, index) {
   list.pop();
 }
 
-const kNodeModulesRE = /^(.*)[\\/]node_modules[\\/]/;
+const kNodeModulesRE = /^(?:.*)[\\/]node_modules[\\/]/;
 
-let getStructuredStack;
+let getStructuredStackImpl;
 
-function isInsideNodeModules() {
-  if (getStructuredStack === undefined) {
+function lazyGetStructuredStack() {
+  if (getStructuredStackImpl === undefined) {
     // Lazy-load to avoid a circular dependency.
     const { runInNewContext } = require('vm');
     // Use `runInNewContext()` to get something tamper-proof and
-    // side-effect-free. Since this is currently only used for a deprecated API,
-    // the perf implications should be okay.
-    getStructuredStack = runInNewContext(`(function() {
+    // side-effect-free. Since this is currently only used for a deprecated API
+    // and module mocking, the perf implications should be okay.
+    getStructuredStackImpl = runInNewContext(`(function() {
       try { Error.stackTraceLimit = Infinity; } catch {}
       return function structuredStack() {
         const e = new Error();
@@ -497,6 +498,16 @@ function isInsideNodeModules() {
     })()`, { overrideStackTrace }, { filename: 'structured-stack' });
   }
 
+  return getStructuredStackImpl;
+}
+
+function getStructuredStack() {
+  const getStructuredStackImpl = lazyGetStructuredStack();
+
+  return getStructuredStackImpl();
+}
+
+function isInsideNodeModules() {
   const stack = getStructuredStack();
 
   // Iterate over all stack frames and look for the first one not coming
@@ -506,20 +517,27 @@ function isInsideNodeModules() {
       const filename = frame.getFileName();
       // If a filename does not start with / or contain \,
       // it's likely from Node.js core.
-      if (RegExpPrototypeExec(/^\/|\\/, filename) === null)
+      if (
+        filename[0] !== '/' &&
+        StringPrototypeIncludes(filename, '\\') === false
+      ) {
         continue;
+      }
       return RegExpPrototypeExec(kNodeModulesRE, filename) !== null;
     }
   }
   return false;
 }
 
-function once(callback) {
+function once(callback, { preserveReturnValue = false } = kEmptyObject) {
   let called = false;
+  let returnValue;
   return function(...args) {
-    if (called) return;
+    if (called) return returnValue;
     called = true;
-    return ReflectApply(callback, this, args);
+    const result = ReflectApply(callback, this, args);
+    returnValue = preserveReturnValue ? result : undefined;
+    return result;
   };
 }
 
@@ -830,6 +848,7 @@ function guessHandleType(fd) {
 
 class WeakReference {
   #weak = null;
+  // eslint-disable-next-line no-unused-private-class-members
   #strong = null;
   #refCount = 0;
   constructor(object) {
@@ -887,6 +906,7 @@ module.exports = {
   getConstructorOf,
   getCWDURL,
   getInternalGlobal,
+  getStructuredStack,
   getSystemErrorMap,
   getSystemErrorName,
   guessHandleType,
