@@ -15,8 +15,8 @@ const {
   ObjectSetPrototypeOf,
   Promise,
   PromisePrototypeThen,
-  PromiseResolve,
   PromiseReject,
+  PromiseResolve,
   SafePromiseAll,
   Symbol,
   SymbolAsyncIterator,
@@ -30,8 +30,8 @@ const {
   AbortError,
   codes: {
     ERR_ILLEGAL_CONSTRUCTOR,
-    ERR_INVALID_ARG_VALUE,
     ERR_INVALID_ARG_TYPE,
+    ERR_INVALID_ARG_VALUE,
     ERR_INVALID_STATE,
     ERR_INVALID_THIS,
     ERR_OUT_OF_RANGE,
@@ -545,12 +545,14 @@ class ReadableStream {
       },
 
       return(error) {
-        return state.current ?
+        started = true;
+        state.current = state.current !== undefined ?
           PromisePrototypeThen(
             state.current,
             () => returnSteps(error),
             () => returnSteps(error)) :
           returnSteps(error);
+        return state.current;
       },
 
       [SymbolAsyncIterator]() { return this; },
@@ -1523,7 +1525,7 @@ function readableStreamPipeTo(
       abortAlgorithm();
       return promise.promise;
     }
-    addAbortListener ??= require('events').addAbortListener;
+    addAbortListener ??= require('internal/events/abort_listener').addAbortListener;
     disposable = addAbortListener(signal, abortAlgorithm);
   }
 
